/****************************************************************************************/
/* CountDown Timer - A simple application for display time-remaining for an event,	*/
/*	such as for cooking, or governing floor-time at a public speaking event.	*/
/*
Usage:
  1. Set the alloted time in the alloted time form box.
  2. Start timing by pressing the Start button.
      You will then see a running display of time remaining and elapsed time.
      The Start button becomes the Stop button.
  3. To Pause the timer, press the Pause button, which then becomes the Resume button.
  4. To Reset the timer, press the Reset button (which then becomes the Start button
      again, much like mechanical stopwatches do).

  When less than 10% time remains, the display turns from green to yellow.
  When time runs out, it turns red and starts blinking.

  A special "Options" button in the upper right corner, allows setting
  the transition time, and other features, as we think to add them.


 Compile on Linux with:
   cc -O -I/usr/X11R6/include -L/usr/X11R6/lib countdown_timer.c \
	-lGLU -lGL -lXmu -lXext -lX11 -lm -o countdown_timer

 Compile on Microsoft with:
   cc countdowntimer.c -lglu32 -lopengl32 -lwinmm -lgdi32 -o countdowntimer.exe

 Depends on OpenToolKit (OTK).
 Get OTK from:  http://otk.sourceforge.net/

 Created by OTK GUI-Builder v0.22

 CountDown-Timer is distributed under BSD License:
 Copyright (c) <2012>, <Skeeter Mon>
 All rights reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.
 THIS SOFTWARE IS PROVIDED BY <copyright holder> ''AS IS'' AND ANY
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 Version 1.0 - Initial Release.

 For updates and more information, see project home page:
   http://sourceforge.net/projects/timercountdown/

 Original Requirements:	 (Why this program was written.)
  We needed a timer display that:
    - Counts backwards down to zero. (Shows Time-remaining)
    - A smaller simultaneous display for elapsed time, would be useful.
    - Should allow full-screen display (to see from a distance), 
	or small (for viewing along side other windows).
    - When it first opens, it should have a dialog window open for
	setting the duration, and any other settings before beginning.
    - Once set up, the time display should occupy the majority of the window, 
	but there should be a small button to Start, Pause, and Reset the timer,
	as well as a button to go back to the settings window, and to exit.
    - The time-display should change color when getting close to timing-out, 
	and it should blink a few times and change to different warning-color
	when the time-duration expires.

    - It would be good if it accepts time in minutes, as well as 
	Days:Hours:Minutes:Seconds, and various subsets there of.
    - It should allow setting the time of transition between warning-colors.
    - The display should be reasonable for short times and long times.
       In other words, don't waste display space showing days and hours, 
	if I am timing a 3-minute egg.
       But allow us to use it for displaying multi-hour or multi-day counts too.
	(It's a computer.  It should be able to automatically optimize its display.)
    - Be programmed in C-language, so everyone can extend it and integrate easily
	with other things.  And so it can run on any machine without installing 
	extra stuff.
    - It should run without needing an internet connection or web browser.
    - Keep it simple, but versatile.
    - There is no need to set any special fonts.  Any normal font will do.
    - Should support option to run an arbitrary command or script when the time
	expires.  Could be used to play a sound file, close stage curtains, etc..

I could not find an application matching all these specifications.
So I went ahead and wrote it.

*/

#include "otk_lib/otk_lib.c"
OtkWidget btn1, label1, hms, window1=0, frmbx1, frmbx2;
OtkWidget Telapsed, Tdisp, panel1;
int ms_disp='m', ls_disp='s';
int warning_set=0, disp_length;

/* All times are internally stored in Seconds. */
double T1, T2, lastT=0.0, last_mark=-1.0;
double allotted_time=60.0*10.0, time_remaining, elapsed_time=0.0, transition_time=60.0;
char duration_string[1024]="10.0";
char transition_string[1024]=" 1.0";


#define WaitingToStart		0	/* Waiting to start.	  	 */
#define RunningDown		1	/* Running (counting down). 	*/ 
#define Paused			2	/* Waiting to re-start.  	*/

int mode=WaitingToStart;


void display_hms( OtkWidget wdg, double dt )
{
 char daywrd[200], hwrd[200], minwrd[200], secswrd[200], word[1024];
 int days, hrs, mins, secs, sgn;

 if (dt < 0.0) { dt = -dt;  strcpy( word, "-" ); } else strcpy( word, "" );
 days = dt / (24.0 * 3600.0);
 dt = dt - days * 24.0 * 3600.0;
 hrs = dt / 3600.0;
 dt = dt - hrs * 3600.0;
 mins = dt / 60;
 secs = dt - 60 * mins;
 if (secs < 10) sprintf(secswrd,"0%d", secs );  else  sprintf(secswrd,"%d", secs );

 switch (ms_disp)
  {
   case 'd':
	sprintf(daywrd,"%d", days );
	strcat( word, daywrd );
	strcat( word, ":" );
	if (hrs < 10) sprintf(hwrd,"0%d", hrs);  else  sprintf(hwrd,"%d", hrs);
   case 'h':
	if (ms_disp == 'h') sprintf(hwrd,"%d", hrs);
	strcat( word, hwrd );
	strcat( word, ":" );
	if (mins < 10) sprintf(minwrd,"0%d", mins);  else  sprintf(minwrd,"%d", mins);
   case 'm':
	if (ms_disp == 'm') sprintf(minwrd,"%d", mins);
	strcat( word, minwrd );
	strcat( word, ":" );
	strcat( word, secswrd );
	Otk_Modify_Text( wdg, word );
  	break;
  }

 if ((wdg == Tdisp) && (strlen( word ) != disp_length))
  {
   float sz, aspct;
   disp_length = strlen( word );
   switch (disp_length)
    {
     case 4:  sz = 32.0;  aspct = 0.65;	break;
     case 5:  sz = 28.0;  aspct = 0.5;	break;
     case 6:  sz = 27.0;  aspct = 0.45;	break;
     case 7:  sz = 24.0;  aspct = 0.4;	break;
     case 8:  sz = 23.0;  aspct = 0.35;	break;
     case 9:  sz = 21.0;  aspct = 0.33;	break;
     case 10:  sz = 20.0;  aspct = 0.3;	break;
     case 11:  sz = 18.5;  aspct = 0.28; break;
     case 12:  sz = 17.5;  aspct = 0.28; break;
     default:
	if (sz < 4)
	 { sz = 33.0;  aspct = 0.7;  }
	else
	 { sz = 16.0;  aspct = 0.25;  }
    }
   Otk_RemoveObject( Tdisp );
   Tdisp = OtkMakeTextLabel( panel1, word, Otk_Green, sz, 2, 5, 18 );
   Otk_Modify_Text_Aspect( Tdisp, aspct );
  }
}


void start( void *data )
{
 switch (mode)
  {
   case WaitingToStart:		/* Start pressed. */
	T1 = otk_report_time();
	printf("Started ...\n");
	Otk_Modify_Text( btn1, "   Pause  " );
	// Otk_Modify_Text_Color( btn1->children, Otk_Red );
	Otk_Modify_Text( label1, "Running ..." );
	Otk_Modify_Text_Color( label1, OtkSetColor( 0.0, 0.2, 0.0 ) );
	mode = RunningDown;
     break;
   case RunningDown:		/* Pause pressed. */
	T2 = otk_report_time();
	Otk_Modify_Text( btn1, "Resume" );
	Otk_Modify_Text( label1, "Paused ..." );
	Otk_Modify_Text_Color( label1, OtkSetColor( 0.2, 0.0, 0.0 ) );
	mode = Paused;
     break;
   case Paused:			/* Resumed. */
	T1 = otk_report_time() - (T2 - T1);
	printf("Resumed ...\n");
	Otk_Modify_Text( btn1, "   Pause  " );
	Otk_Modify_Text( label1, "Running ..." );
	Otk_Modify_Text_Color( label1, OtkSetColor( 0.0, 0.2, 0.0 ) );
	mode = RunningDown;
     break;
  } 
}


void reset( void *data )
{
 mode = WaitingToStart;
 Otk_Modify_Text( btn1, "  Start " );
 Otk_Modify_Text( label1, "Ready to start . . ." );
 Otk_Modify_Text_Color( label1, Otk_DarkGray );
 display_hms( Telapsed, 0.0 ); 
 display_hms( Tdisp, allotted_time );
 warning_set = 0;
 Otk_Modify_Text_Color( Tdisp, Otk_Green );
}


void quit( void *data ) { exit(0); }


void cancel( void *x )
{ 
 Otk_RemoveObject( window1 );
 window1 = 0;
}


void decode_tvalue( char *text, double *tvalue )
{
 char line[1024], word[1024];
 double day, hour, minute, second;
 int k;

 /*Eliminate any leading or trailing white-space.*/
 k = strlen(text)-1;
 while ((k > 0) && ((text[k] == ' ') || (text[k] == '\t'))) { text[k] = '\0';  k--; }
 while ((text[0] == ' ') || (text[0] == '\t')) { strcpy( text, &(text[1]) ); }

 if ((strstr( text, ":" ) != 0) || (strstr( text, ";" ) != 0))	/* Accept ";" as common typo. */
  { /*Expect d:h:m:s, or h:m:s, or m:s.*/
    strcpy( line, text );
    Otk_next_word( line, word, ":;");
    if (sscanf(word,"%lf", &day) != 1)
     { printf( "Bad time value '%s'\n", text );  return; }
    Otk_next_word( line, word, ":;");
    if (sscanf(word,"%lf", &hour) != 1)
     { printf( "Bad time value '%s'\n", text );  return; }
    Otk_next_word( line, word, ":;");
    if (word[0] == '\0')
     { /*m:s:*/
       *tvalue = 60.0 * day + hour;
       return;
     }
    if (sscanf(word,"%lf", &minute) != 1)
     { printf( "Bad time value '%s'\n", text );  return; }
    Otk_next_word( line, word, ":;");
    if (word[0] == '\0')
     { /*h:m:s:*/
       *tvalue = 3600.0 * day + 60.0 * hour + minute;
       return;
     }
    if (sscanf(word,"%lf", &second) != 1)
     { printf( "Bad time value '%s'\n", text );  return; }
    *tvalue = 24 * 3600.0 * day + 3600.0 * hour + 60.0 * minute + second;
  }
 else
  { /*Plain minutes value.*/
   if (sscanf(text,"%lf", tvalue) != 1)
    { printf( "Bad floating-point minutes value '%s'\n", text );  return; }
   *tvalue = 60.0 * *tvalue;
  }
}


void accept_allotted_time( void *x )
{
 char text[100];
 double transition_percentage=0.15;
 int k;

 Otk_Get_Text( frmbx1, duration_string, 100 );
 decode_tvalue( duration_string, &allotted_time );

 Otk_Get_Text( frmbx2, transition_string, 100 );
 decode_tvalue( transition_string, &transition_time );

 time_remaining = allotted_time - elapsed_time;

 if (allotted_time >= 24.0 * 3600.0)  ms_disp = 'd';  else
 if (allotted_time >= 3600.0)  ms_disp = 'h';
 else ms_disp = 'm';

 if (mode == WaitingToStart)
  display_hms( Tdisp, allotted_time );
 else
  display_hms( Tdisp, allotted_time - (lastT - T1) ); 

 printf("Duration = %g minutes, Warning at %g minutes.\n", allotted_time/60.0, transition_time/60.0 );
 cancel(0);
}

void accept_allotted_time2( char *s, void *x )
{
 accept_allotted_time( x );
}


void killed_window( void *x )
{
 window1 = 0;
}

void quit_out( void *data ) 
{
  Otk_RemoveObject( window1 );
  window1 = OtkMakeWindow( Otk_Recessed, Otk_Blue, Otk_LightGray, 10, 55, 75, 30 );
  OtkMakeButton( window1, 10, 15.0, 35, 70, "Yes, Quit ", quit, 0 );
  OtkMakeButton( window1, 60, 15.0, 35, 70, "No, Cancel ", cancel, 0 );
  Otk_RegisterWindowKillEventFunction( window1, killed_window, 0 );
}



void disp_units( void *data )
{
 char *strng=(char *)data;

 if (strcmp( strng, "Lead_days") == 0)     ms_disp = 'd';	else
 if (strcmp( strng, "Lead_hours") == 0)    ms_disp = 'h';	else
 if (strcmp( strng, "Lead_minutes") == 0)  ms_disp = 'm';	else
 if (strcmp( strng, "LS_minutes") == 0)    ls_disp = 'm';	else
 if (strcmp( strng, "LS_seconds") == 0)    ls_disp = 's';
 else printf("Error\n");
 // printf("Set disp units to %c %c\n", ms_disp, ls_disp );

 ls_disp = 's';
 if (allotted_time >= 24.0 * 3600.0)  ms_disp = 'd';  else
 if (allotted_time >= 3600.0)  ms_disp = 'h';
 else ms_disp = 'm';
}




void setup_window( void *data )
{
  float x1=10, y1=21, x2=55, y2=73, wdth, hgt=10;
  OtkWidget rad[10];
  char value[100];

  if (window1 != 0) Otk_RemoveObject( window1 );
  Otk_Set_Default_Border_Thickness( 0.25 );
  window1 = OtkMakeWindow( Otk_Recessed, Otk_Blue, Otk_LightGray, 10, 5, 80, 90 );
  Otk_Set_Default_Border_Thickness( 1.0 );

  Otk_Set_Text_Aspect( 0.7 );
  OtkMakeTextLabel( window1, "Set Duration =", Otk_Black, 2.5, 1.0, 2, 7 );
  OtkMakeTextLabel( window1, "Minutes", Otk_Black, 1.6, 1.0, y2, 6.5 );
  OtkMakeTextLabel( window1, "(Or D:H:M:S)", Otk_Black, 1.6, 1.0, y2, 12 );
  OtkMakeTextLabel( window1, "Change Color at Time Remaining =", Otk_Black, 2.5, 1.0, 2, y1+2 );
  OtkMakeTextLabel( window1, "Minutes", Otk_Black, 1.6, 1.0, y2, y1+10.5 );
  OtkMakeTextLabel( window1, "(Or D:H:M:S)", Otk_Black, 1.6, 1.0, y2, y1+15 );
  Otk_Set_Text_Aspect( 1.0 );

  frmbx1 = OtkMakeTextFormBox( window1, duration_string, 12, 34, 6, 35, hgt, accept_allotted_time2, 0 );
  Otk_Center_Justify_Formbox( frmbx1 );

  Otk_Add_Line( window1, Otk_Blue, 1.0, 0, y1-0.5,  100, y1-0.5 );

  frmbx2 = OtkMakeTextFormBox( window1, transition_string, 8, 34, y1+10, 35, hgt, accept_allotted_time2, 0 );
  Otk_Set_Text_Aspect( 1.0 );
  Otk_Center_Justify_Formbox( frmbx2 );

  y1 = y1 + hgt + 13;
  Otk_Add_Line( window1, Otk_Blue, 1.0, 0, y1,  100, y1 );

  y1 = y1 + 1;  wdth=6, hgt=4;
  Otk_Set_Text_Aspect( 0.7 );
  OtkMakeTextLabel( window1, "Time Units Displayed:", Otk_Black, 2.5, 1.0, 2, y1 + 1 );
  y1 = y1 + 9;
  OtkMakeTextLabel( window1, "(Leading Unit)", Otk_Black, 1.8, 1.0, 8,  y1 );
  OtkMakeTextLabel( window1, "(Least Significant Unit)", Otk_Black, 1.8, 1.0, x2-1, y1 );
  y1 = y1 + 6;
  rad[0] = OtkMakeRadioButton( window1, x1, y1, wdth, hgt, disp_units, "Lead_days" ); 
  OtkMakeTextLabel( window1, "Days", Otk_Black, 1.5, 1.0, x1+wdth, y1 );
  rad[3] = OtkMakeRadioButton( window1, x2, y1, wdth, hgt, disp_units, "LS_minutes" ); 
  OtkMakeTextLabel( window1, "Minutes", Otk_Black, 1.5, 1.0, x2+wdth, y1 );
  y1 = y1 + 6;
  rad[1] = OtkMakeRadioButton( rad[0], x1, y1, wdth, hgt, disp_units, "Lead_hours" ); 
  OtkMakeTextLabel( window1, "Hours", Otk_Black, 1.5, 1.0, x1+wdth, y1 );
  rad[4] = OtkMakeRadioButton( rad[3], x2, y1, wdth, hgt, disp_units, "LS_seconds" ); 
  OtkMakeTextLabel( window1, "Seconds", Otk_Black, 1.5, 1.0, x2+wdth, y1 );
  y1 = y1 + 6;
  rad[2] = OtkMakeRadioButton( rad[0], x1, y1, wdth, hgt, disp_units, "Lead_minutes" ); 
  OtkMakeTextLabel( window1, "Minutes", Otk_Black, 1.5, 1.0, x1+wdth, y1 );

  switch (ms_disp)
   {
    case 'd':  Otk_SetRadioButton( rad[0] );	break;
    case 'h':  Otk_SetRadioButton( rad[1] );	break;
    case 'm':  Otk_SetRadioButton( rad[2] );	break;
   }
  if (ls_disp == 'm')
   Otk_SetRadioButton( rad[3] );
  else
   Otk_SetRadioButton( rad[4] );
  Otk_Set_Text_Aspect( 1.0 );

  y1 = 84.5;
  Otk_Add_Line( window1, Otk_Blue, 1.0, 0, y1-3,  100, y1-3 );
  Otk_Set_Default_Button_BorderThickness( 1.75 );
  OtkMakeButton( window1, 7, y1, 20, 10, "Quit ? ", quit_out, 0 );
  OtkMakeButton( window1, 65, y1, 24, 10, " OK  ", accept_allotted_time, 0 );
  Otk_Set_Default_Button_BorderThickness( 1.0 );

  Otk_RegisterWindowKillEventFunction( window1, killed_window, 0 );
}



main( int argc, char **argv )
{
  char *action=0;
  double dt, y1=12;
  int argk=1;

  while (argk < argc)
   {
    if (strcmp( argv[argk], "-action" ) == 0)
     { /* Get optional action command, to be performed when timer expires. */
	argk++;
	if (argk == argc)
	 { printf("Error: Missing command after '%s'.\n", argv[argk-1] );  exit(1); }
	action = strdup( argv[ argk ] );
     }
    argk++;
   }

  Otk_Set_Window_Name( "CountDown Timer" );
  OtkInitWindow( 510, 360, argc, argv );
  
  Otk_Set_Default_Button_BorderThickness( 1.75 );
  btn1 = OtkMakeButton( OtkOuterWindow, 2, 2, 25, y1-4, "  Start ", start, 0 );
  Otk_Modify_Text_Color( btn1->children, OtkSetColor( 0.0, 0.2, 0.0 ) );
  OtkMakeButton( OtkOuterWindow, 84, 2.5, 14, y1-5, "Options", setup_window, 0 );
  label1 = OtkMakeTextLabel( OtkOuterWindow, "Ready to start . . .", Otk_DarkGray, 2.0, 1, 35, 2 );
  OtkMakeTextLabel( OtkOuterWindow, "Elapsed Time:", Otk_Blue, 2.0, 1, 35, 7 );
  Telapsed = OtkMakeTextLabel( OtkOuterWindow, "0.0", Otk_Blue, 2.0, 1, 60, 7 );
  Otk_Add_Line( OtkOuterWindow, Otk_Blue, 2.0, 0, y1,  100, y1 );

  panel1 = OtkMakePanel( OtkOuterWindow, Otk_Flat , Otk_Black, 0, y1+0.5, 100, 100 - (y1 + 0.5) ); 
  OtkMakeTextLabel( panel1, "Time Remaining:", Otk_White, 3.0, 1, 2, 3 );
  Otk_Set_Default_Button_BorderThickness( 3.0 );
  OtkMakeButton( panel1, 90, 2, 9, 7, "Reset ", reset, 0 );
  Otk_Set_Default_Button_BorderThickness( 1.0 );

  Tdisp = OtkMakeTextLabel( panel1, "10:00", Otk_Green, 28.0, 2, 5, 18 );
  Otk_Modify_Text_Aspect( Tdisp, 0.5 );
  disp_length = 5;

	#if (0)
		while (1)
		{ char word[1000]; float sz, aspct;
		 printf("Enter string ? ");
		 scanf("%s", word);
		 do
		 {
		  printf("Enter sz + aspct ? ");
		  scanf("%f", &sz); scanf("%f", &aspct);
		  if (sz > 0.0)
		   {
		    Otk_RemoveObject( Tdisp );  
		    Tdisp = OtkMakeTextLabel( panel1, word, Otk_Green, sz, 2, 5, 18 );
		    Otk_Modify_Text_Aspect( Tdisp, aspct );
		    OtkUpdateCheck();
		   }
		 } while (sz > 0.0);
		}
	#endif

  setup_window(0);

  // OtkMainLoop();
  while (1)
   {
    if (mode == RunningDown)
     {
      if (otk_report_time() - lastT >= 0.25)
       {
	lastT = otk_report_time();
	dt = lastT - T1;	/* Total Elapsed-time. */
	display_hms( Telapsed, dt ); 
	display_hms( Tdisp, allotted_time - dt ); 

	if ((allotted_time - dt <= 0.0) && (warning_set < 2))
	 { /*Time-expires*/
	  Otk_Modify_Text_Color( Tdisp, Otk_Red );
	  Otk_Set_Button_Color( panel1, Otk_White );
	  warning_set = 2;
	  if (action != 0)
	   {
	    printf("Issuing: '%s'\n", action );
	    system( action );
	    action = 0;
	   }
	 } 
	else
	if ((allotted_time - dt < transition_time) && (warning_set == 0))
	 { /*Time-is-nearly-out*/
	  Otk_Modify_Text_Color( Tdisp, Otk_Yellow );
	  warning_set = 1;
	 } 
	else
	if ((warning_set >= 2) && (warning_set < 7 ))
	 { /*Rapid-blinks-after-expiration*/
	  if (warning_set % 2 == 0)
	   Otk_Set_Button_Color( panel1, Otk_Black );
	  else
	   Otk_Set_Button_Color( panel1, Otk_Yellow );
	  warning_set++;
	 }

       }
     }
    OtkUpdateCheck();
    otk_sleep( 0.1 );
   }
}
